/*
 * Copyright (C) 2003 Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 */

#include <stdio.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include "libppm.h"

#define WIDTH 500
#define HEIGHT 500

struct box {
	int x1, y1, x2, y2;
};

static int init(int argc, char **argv,
		struct ppm_image *p1, struct ppm_image *p2);
static int is_dark(struct ppm_image *ppm, int x, int y);
static int is_cut(struct ppm_image *ppm,
		int x1, int y1, int x2, int y2, int dx, int dy);
static void find_box(struct ppm_image *ppm, int x, int y, struct box *ret);

static gint expose_cb(GtkWidget *widget, GdkEventExpose *event,
		struct ppm_image *ppm);
static gint click_cb(GtkWidget *widget, GdkEventButton *event,
		struct ppm_image *ppm);

static int crop = FALSE;
static struct box bound;
static float scale_x, scale_y;
static const char *output_file_pattern;
static int output_number = 1;

int main(int argc, char **argv) {
	struct ppm_image ppm, small;
	GtkWidget *window, *da;

	gtk_init(&argc, &argv);
	gdk_rgb_init();

	if (init(argc, argv, &ppm, &small) == -1) return 1;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	da = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(da), WIDTH, HEIGHT);
	scale_x = (float)ppm.width/WIDTH;
	scale_y = (float)ppm.height/HEIGHT;
	gtk_container_add(GTK_CONTAINER(window), da);
	gtk_signal_connect(GTK_OBJECT(da), "expose_event",
		GTK_SIGNAL_FUNC(expose_cb), &small);
	gtk_signal_connect(GTK_OBJECT(da), "button_press_event",
		GTK_SIGNAL_FUNC(click_cb), &ppm);
	gtk_signal_connect(GTK_OBJECT(window), "destroy", gtk_main_quit, NULL);
	gtk_widget_set_events(da, GDK_EXPOSURE_MASK|GDK_BUTTON_PRESS_MASK);
	gtk_widget_show_all(window);

	gtk_main();

	return 0;
}
static int is_dark(struct ppm_image *ppm, int x, int y) {
	PEL *p = &PPM_PIXEL(*ppm, x, y, 0);
	return (p[0] + p[1] + p[2] < 200*3);
}

static int is_cut(struct ppm_image *ppm,
		int x1, int y1, int x2, int y2, int dx, int dy) {
	int x=x1, y=y1, dark=0, count=0, last=0;

	do {
		if (is_dark(ppm, x, y))
			dark++;
		count++;
		if (x == x2 && y == y2) last = 1;
		x+=dx;
		y+=dy;
	} while (!last);

	return dark==0 || count/dark > 100;
}

static void find_box(struct ppm_image *ppm, int x, int y, struct box *ret) {
	int x1=x, y1=y, x2=x, y2=y, clear=0;

	while (1) {
		if (y1 <= 0 || is_cut(ppm, x1, y1, x2, y1, 1, 0)) {
			if (++clear == 4) break;
		}
		else { clear=0; y1--; }

		if (y2 >= ppm->height-1 || is_cut(ppm, x1, y2, x2, y2, 1, 0)) {
			if (++clear == 4) break;
		}
		else { clear=0; y2++; }

		if (x1 <= 0 || is_cut(ppm, x1, y1, x1, y2, 0, 1)) {
			if (++clear == 4) break;
		}
		else { clear=0; x1--; }

		if (x2 >= ppm->width-1 || is_cut(ppm, x2, y1, x2, y2, 0, 1)) {
			if (++clear == 4) break;
		}
		else { clear=0; x2++; }
	}

	ret->x1 = x1;
	ret->y1 = y1;
	ret->x2 = x2;
	ret->y2 = y2;
}

static int init(int argc, char **argv,
		struct ppm_image *p1, struct ppm_image *p2) {
	FILE *inp;
	char *file_check;
	char *test1, *test2;
	struct stat ign;

	if (argc != 3) {
		fprintf(stderr, "Usage:\n  %s input.ppm output%%d.ppm\n\n", argv[0]);
		return -1;
	}
	output_file_pattern = argv[2];
	test1 = g_strdup_printf(output_file_pattern, 1);
	test2 = g_strdup_printf(output_file_pattern, 2);
	if (!strcmp(test1, test2)) {
		fprintf(stderr, "Invalid output pattern \"%s\"\n", output_file_pattern);
		fprintf(stderr, "Output pattern must have a %%d in it.\n\n");
		return -1;
	}
	g_free(test1);
	g_free(test2);

	inp = fopen(argv[1], "r");
	if (!inp) {
		perror(argv[1]);
		return -1;
	}
	ppminit(p1);
	if (ppmmap(inp, p1) == -1) {
		fprintf(stderr, "Error reading PPM.\n");
		return 1;
	}
	fclose(inp);

	ppminit(p2);
	ppmcopy(p2, p1);
	ppmscale(p2, WIDTH, HEIGHT);

	file_check = g_strdup_printf(output_file_pattern, output_number);
	while (stat(file_check, &ign) != -1) {
		g_free(file_check);
		file_check = g_strdup_printf(output_file_pattern, ++output_number);
	}
	g_free(file_check);

	return 0;
}

static gint expose_cb(GtkWidget *widget, GdkEventExpose *event,
		struct ppm_image *ppm) {
	if (crop) {
		int x1, y1, x2, y2, i;
		char *pixels = g_new(char, event->area.height*WIDTH*3);
		memcpy(pixels, ppm->data + 3*WIDTH*event->area.y,
			3*WIDTH*event->area.height);
		x1 = bound.x1 / scale_x;
		y1 = bound.y1 / scale_y;
		x2 = bound.x2 / scale_x;
		y2 = bound.y2 / scale_y;
		for (i=x1; i<=x2; i++) {
			pixels[(y1*WIDTH+i)*3] = 255;
			pixels[(y1*WIDTH+i)*3+1] = 0;
			pixels[(y1*WIDTH+i)*3+2] = 255;
			pixels[(y2*WIDTH+i)*3] = 255;
			pixels[(y2*WIDTH+i)*3+1] = 0;
			pixels[(y2*WIDTH+i)*3+2] = 255;
		}
		for (i=y1; i<=y2; i++) {
			pixels[(i*WIDTH+x1)*3] = 255;
			pixels[(i*WIDTH+x1)*3+1] = 0;
			pixels[(i*WIDTH+x1)*3+2] = 255;
			pixels[(i*WIDTH+x2)*3] = 255;
			pixels[(i*WIDTH+x2)*3+1] = 0;
			pixels[(i*WIDTH+x2)*3+2] = 255;
		}
		gdk_draw_rgb_image(widget->window, widget->style->black_gc,
			0, event->area.y, WIDTH, event->area.height, GDK_RGB_DITHER_NONE,
			pixels, WIDTH*3);
		g_free(pixels);
	}
	else
		gdk_draw_rgb_image(widget->window, widget->style->black_gc,
			0, event->area.y, WIDTH, event->area.height, GDK_RGB_DITHER_NONE,
			ppm->data + 3*WIDTH*event->area.y, WIDTH*3);

	return TRUE;
}

static gint click_cb(GtkWidget *widget, GdkEventButton *event,
		struct ppm_image *ppm) {
	int x, y;
	char *filename;

	x = event->x * ppm->width / WIDTH;
	y = event->y * ppm->height / HEIGHT;

	if (crop && x > bound.x1 && x < bound.x2 && y > bound.y1 && y < bound.y2) {
		struct ppm_image newppm;
		FILE *fp;

		filename = g_strdup_printf(output_file_pattern, output_number++);
		fp = fopen(filename, "w");
		if (!fp) {
			perror("filename");
			crop = FALSE;
			gtk_widget_queue_draw(widget);
			return TRUE;
		}
		printf("writing to \"%s\"\n", filename);
		g_free(filename);

		ppminit(&newppm);
		ppmalloc(&newppm, bound.x2-bound.x1-1, bound.y2-bound.y1-1, 0);
		for (y=bound.y1+1; y<bound.y2; y++)
			memcpy(&PPM_PIXEL(newppm, 0, y-bound.y1-1, 0),
				&PPM_PIXEL(*ppm, bound.x1+1, y, 0),
				newppm.width*3);
		ppmwrite(fp, &newppm);
		ppmdestroy(&newppm);

		fclose(fp);
		crop = FALSE;
		gtk_widget_queue_draw(widget);
		return TRUE;
	}

	find_box(ppm, x, y, &bound);

	printf("click: %d,%d => %d,%d bounded by %d,%d - %d,%d\n",
		(int)event->x, (int)event->y, x, y, bound.x1, bound.y1, bound.x2, bound.y2);

	crop = TRUE;
	gtk_widget_queue_draw(widget);

	return TRUE;
}
